from pydependabot.consts import NOT_POPULATED

class Logger(object):
    def __init__(self, print_debug_log):
        self.print_debug_log = print_debug_log

    def debug(self, line):
        if self.print_debug_log:
            print(f"[DEBUG] {line}")
    

def get_if_possible(json, element):
    """  returns the value of the dict key, according to its path """
    # if the key doesnt exists NOT_POPULATED will be returned.
    # i.e. get_if_possible("key1.key2", {"key1": {"key2": "abc"}} )
    try:
        keys = element.split('.')
        rv = json
        for key in keys:
            rv = rv[key]
        return rv

    except KeyError:
        return NOT_POPULATED
    except TypeError:
        return NOT_POPULATED




from datetime import datetime
from packaging import version

from pydependabot.utils import get_if_possible
from pydependabot.consts import NOT_POPULATED


class securityVulnerability(object):
	def __init__(self):
		# we list here the properties of security vulnerability for autocompletion in IDE. that's the only reason.
		self.severity = None
		self.first_patched_version = None
		self.package_ecosystem = None
		self.package_name = None
		self.updated_at = None
		self.vulnerable_version_range = None
		self.advisory_summary = None
		self.advisory_cvss_score = None

class Dismisser(object):
	def __init__(self):
		# we list here the properties of security vulnerability for autocompletion in IDE. that's the only reason.
		self.email = None
		self.name = None
		self.login = None
		self.url = None
		return 


class RepositoryVulnerablityAlert(object):
	def __init__(self, node, repository_owner, repository_name):
		self.node = node
		self.repository_owner = repository_owner
		self.repository_name = repository_name

	@property
	def securityVulnerability(self):
		security_vulnerability = securityVulnerability()
		security_vulnerability.severity = get_if_possible(self.node, "securityVulnerability.severity")
		security_vulnerability.package_ecosystem = get_if_possible(self.node, "securityVulnerability.package.ecosystem")
		security_vulnerability.package_name = get_if_possible(self.node, "securityVulnerability.package.name")
		security_vulnerability.vulnerable_version_range = get_if_possible(self.node, "securityVulnerability.vulnerableVersionRange")
		security_vulnerability.advisory_cvss_score = get_if_possible(self.node, "securityVulnerability.advisory.cvss.score")
		security_vulnerability.advisory_summary = get_if_possible(self.node, "securityVulnerability.advisory.summary")
		
		first_patched_version = get_if_possible(self.node, "securityVulnerability.firstPatchedVersion.identifier") 
		security_vulnerability.first_patched_version = version.parse(first_patched_version) if first_patched_version is not NOT_POPULATED else NOT_POPULATED
		
		updated_at = get_if_possible(self.node, "securityVulnerability.updatedAt") 
		security_vulnerability.updated_at = datetime.strptime(updated_at, "%Y-%m-%dT%H:%M:%SZ") if updated_at is not NOT_POPULATED else NOT_POPULATED
		
		return security_vulnerability

	@property
	def created_at(self):
		created_at = get_if_possible(self.node, "createdAt")
		return datetime.strptime(created_at, "%Y-%m-%dT%H:%M:%SZ") if created_at is not NOT_POPULATED else NOT_POPULATED
		
	@property
	def dismissed_at(self):
		dismissed_at = get_if_possible(self.node, "dismissedAt")
		return datetime.strptime(dismissed_at, "%Y-%m-%dT%H:%M:%SZ") if dismissed_at is not NOT_POPULATED else NOT_POPULATED

	@property
	def dismisser(self):
		dismisser = Dismisser()
		dismisser.name = get_if_possible(self.node, "dismisser.name")
		dismisser.login = get_if_possible(self.node, "dismisser.login")
		dismisser.email = get_if_possible(self.node, "dismisser.email")
		dismisser.url = get_if_possible(self.node, "dismisser.url")
		return dismisser

	
	@property
	def state(self):
		return get_if_possible(self.node, "state")
		
	@property
	def vulnerable_manifest_path(self):
		return get_if_possible(self.node, "vulnerableManifestPath")
		

	@property
	def vulnerable_requirements(self):
		return get_if_possible(self.node, "vulnerableRequirements")

	@property
	def vulnerable_manifest_filename(self):
		return get_if_possible(self.node, "vulnerableManifestFilename")

	@property
	def dismiss_comment(self):
		return get_if_possible(self.node, "dismissComment")

	@property
	def dismiss_reason(self):
		return get_if_possible(self.node, "dismissReason")
	
	@property
	def fixed_at(self):
		fixed_at = get_if_possible(self.node, "fixedAt")
		return datetime.strptime(fixed_at, "%Y-%m-%dT%H:%M:%SZ") if fixed_at is not NOT_POPULATED else NOT_POPULATED
	
	@property
	def number(self):
		return get_if_possible(self.node, "number")
	
	def __repr__(self):
		return f"""(Repository:{self.repository_owner}/{self.repository_name}, Package:{self.securityVulnerability.package_name}:{self.securityVulnerability.severity}-severity)"""




class RepositoryVulnerablityAlertQuery(object):
	createdAtField = "createdAt"
	securityVulnerabilityField = """ securityVulnerability { 
										severity,
										firstPatchedVersion  { identifier  },
										package { ecosystem, name },
										updatedAt,
									    vulnerableVersionRange,
										advisory { summary, cvss { score } }, 
									}  """

	dismissedAtField = "dismissedAt"
	dismissCommentField = "dismissComment"
	dismissReasonField = "dismissReason"
	dismisserField = "dismisser { email, name, login, url }" # requires  ['user:email', 'read:user'] permissions
	fixedAtField = "fixedAt"
	numberField = "number"
	stateField = "state"
	vulnerableManifestFilenameField = "vulnerableManifestFilename"
	vulnerableManifestPathField = "vulnerableManifestPath"
	vulnerableRequirementsField =  "vulnerableRequirements"




from gql import gql, Client
from gql.transport.requests import RequestsHTTPTransport
from pydependabot.repository_vulnerablity_alert_query import RepositoryVulnerablityAlertQuery
from pydependabot.repository_vulnerablity_alert_node import RepositoryVulnerablityAlert
from pydependabot.utils import Logger


class Dependabot():
	def __init__(self, graphql_endpoint, auth_token, options={}):
		"""
		graphql_endpoint: The endpoint the communication will be directed to. For github it's usually https://api.github.com/graphql
		auth_token: Will be used to authenticate with the graphql endpoint, must have sufficient permissions
		options: mainly for debugging. accepts
		        - 'print_debug_log' that will print more logs to help debugging
				- 'client' that will be used to replace the graphql client being used, mainly for testing or special cases.

		"""
		self.logger = Logger(print_debug_log=options.get("print_debug_log"))
		self.logger.debug("Debug mode is on")


		if options.get("client") is None:
			transport = RequestsHTTPTransport( url=graphql_endpoint, headers = {"Authorization": f"bearer {auth_token}"} )
			self.client = Client(transport=transport)
		else:
			self.client = options.get("client")

		
		


	def get_security_alerts(self, repository_owner: str, repository_name: str,
							fields=[RepositoryVulnerablityAlertQuery.securityVulnerabilityField, RepositoryVulnerablityAlertQuery.createdAtField]):

		# for more information please read the official docs - https://github.com/or-elias/pyDependabot
		#
		# Input parameters:
		#	- repository_owner - sometimes refered to as the organization of the repository. i.e. microsoft
		#   - repository_name - the full name of the repo. i.e. vscode
		#   - fields - not mandatory. a list of consts from RepositoryVulnerablityAlertQuery class. represents the data we ask from the server.
		# 
		# Output - 
		#	for now the function returns a list of raw nodes as returned from the server. 
		#	also, if we received a graphql error from github (for example, NOT FOUND), we let the gql.transport.exceptions.TransportQueryError excption to be raised.
		#
		if len(fields) == 0:
			raise Exception("GraphQL query must fetch at least one field. Got empty list of fields")
		raw_nodes_from_dependabot = self._query_api_for_repositories_vulnerablities(repository_owner, repository_name, fields)
		return [RepositoryVulnerablityAlert(node, repository_owner, repository_name) for node in raw_nodes_from_dependabot]
		



	def _query_api_for_repositories_vulnerablities(self, repository_owner, repository_name, fields, after=None, current_results=[]) -> list[dict]:
		# A recursive function that returns a list of raw nodes representing vulnerablities alerts from github.
		# A node is a github graphql object as shown here https://docs.github.com/en/graphql/reference/interfaces#node. we fetch it's RepositoryVulnerability fields.
		# Note: This function doesnt have recursive depth limitation since we figured a single repository will probably wont have thousands of alerts.
		#
		
		q = f"""
			query GetRepositoryDependabotAlerts ($name: String!, $owner: String!, $after: String) {{ 
					repository( name: $name, owner: $owner) {{
																 
						vulnerabilityAlerts(first: 99, after: $after) {{  
							pageInfo {{ hasNextPage, endCursor }} ,
							nodes{{ {','.join(fields)} }} 
						}}
					}} 	
			}}
		"""
		vars = {"name": repository_name, "owner":repository_owner, "after": after}
		results = self.client.execute(gql(q), variable_values=vars)
		# Because of the nature of Graphql, I can be sure that I will always get 'endCursor' and 'node' fields in the response, even if they are empty.
		end_cursor = results.get('repository').get('vulnerabilityAlerts').get('pageInfo').get('endCursor')
		nodes = results.get('repository').get('vulnerabilityAlerts').get('nodes')
		
		if end_cursor is None:
			return current_results+nodes
		else:
			return self._query_api_for_repositories_vulnerablities(repository_owner, repository_name, fields, after=end_cursor, current_results=current_results+nodes)







from setuptools import setup, find_packages


DESCRIPTION = "A python library to query and manage github's dependabot alerts"


# Setting up
setup(
    name="pydependabot",
    version='0.8',
    author="Or1337 (Or Elias)",
    author_email="<orelias.tm@gmail.com>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=open("README.md").read(),
    packages=find_packages(),
    install_requires=['gql', 'packaging'],
    keywords=['python', 'Dependabot', 'github', 'security', 'open source security', "pyDependabot", "pydependabot", "py-dependabot"],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)